cdef void scheduler_init(scheduler_t *scheduler, unsigned short num_workers=sysconf(_SC_NPROCESSORS_ONLN)) nogil:
  memset(scheduler, 0, sizeof(scheduler[0]))

  scheduler.num_workers = num_workers
  scheduler.is_done = False
  scheduler.future_async_jobs = False

  pthread_mutex_init(&scheduler.mutex, NULL)
  scheduler.deque = new coro_deque_t()

  sem_init(&scheduler.num_sleeping_workers, 0, 0)
  sem_init(&scheduler.spin_loop_workers, 0, 1)
  
  scheduler.workers = <worker_t *>calloc(scheduler.num_workers, sizeof(worker_t))
  if not scheduler.workers:
    fprintf(stderr, "Could not allocate memory for the workers")
    exit(-1)

  if pthread_barrier_init(&scheduler.barrier, NULL, scheduler.num_workers + 1):
    fprintf(stderr, "Could not allocate memory for the barrier")
    exit(-1)

  cdef unsigned short i
  for i in range(scheduler.num_workers):
    worker_init(&scheduler.workers[i], scheduler)

cdef void scheduler_coro_add(scheduler_t *scheduler, coro_function_t task, void *arg=NULL) nogil:
  pthread_mutex_lock(&scheduler.mutex)
  scheduler.deque.push_back(coro_new(NULL, task, arg))
  pthread_mutex_unlock(&scheduler.mutex)

cdef void scheduler_run(scheduler_t *scheduler) nogil:
  # Wait for all workers and the main thread to be ready.
  pthread_barrier_wait(&scheduler.barrier)

  cdef int num_sleeping_workers
  while not scheduler.is_done:
    sem_getvalue(&scheduler.num_sleeping_workers, &num_sleeping_workers)
    if num_sleeping_workers == scheduler.num_workers and not scheduler.future_async_jobs:
      scheduler.is_done = True
    usleep(1000000) # 1 ms (arbitrary)

cdef void *scheduler_run_thread(void *scheduler) nogil:
  scheduler_run(<scheduler_t*> scheduler)
  pthread_exit(NULL)

cdef void async_scheduler_run(scheduler_t *scheduler) nogil:

  scheduler.future_async_jobs = True
  cdef pthread_attr_t attr
  if pthread_attr_init(&attr):
    fprintf(stderr, "pthread_attr_init()\n")
    exit(-1)

  if pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE):
    fprintf(stderr, "pthread_attr_setdetachstate()\n")
    exit(-1)

  if pthread_create(&scheduler.thread, &attr, scheduler_run_thread, scheduler):
    fprintf(stderr, "pthread_create()\n")
    exit(-1)

  if pthread_attr_destroy(&attr):
    fprintf(stderr, "pthread_attr_destroy()\n")
    exit(-1)

cdef void async_scheduler_destroy(scheduler_t *scheduler) nogil:

  scheduler.future_async_jobs = False
  while not scheduler.is_done:
    usleep(1000000) # 1 ms (arbitrary)

  scheduler_destroy(scheduler)
  pthread_join(scheduler.thread, NULL)

cdef void scheduler_destroy(scheduler_t *scheduler) nogil:
  cdef unsigned short i
  for i in range(scheduler.num_workers):
    pthread_join(scheduler.workers[i].thread, NULL)
    worker_destroy(&scheduler.workers[i])

  free(scheduler.workers)

  sem_destroy(&scheduler.num_sleeping_workers)
  sem_destroy(&scheduler.spin_loop_workers)

  pthread_barrier_destroy(&scheduler.barrier)

  pthread_mutex_destroy(&scheduler.mutex)
  del scheduler.deque

cdef void worker_init(worker_t *worker, scheduler_t *scheduler) nogil:
  memset(worker, 0, sizeof(worker[0]))

  worker.scheduler = scheduler

  worker.deque = new coro_deque_t()
  worker.bypass_global_queue = 0

  if not worker.deque:
    fprintf(stderr, "Could not allocate memory for the deque\n")
    exit(-1)

  pthread_mutex_init(&worker.mutex, NULL)

  cdef pthread_attr_t attr
  if pthread_attr_init(&attr):
    fprintf(stderr, "pthread_attr_init()\n")
    exit(-1)
  
  if pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE):
    fprintf(stderr, "pthread_attr_setdetachstate()\n")
    exit(-1)

  if pthread_create(&worker.thread, &attr, worker_run, worker):
    fprintf(stderr, "pthread_create()\n")
    exit(-1)

  if pthread_attr_destroy(&attr):
    fprintf(stderr, "pthread_attr_destroy()\n")
    exit(-1)

cdef void *worker_run(void *arg) nogil:
  cdef:
    worker_t *worker = <worker_t *>arg
    coro_t *coroutine = NULL
    bint was_waiting = False
    bint has_sem = False

  # Wait for all workers and the main thread to be ready.
  pthread_barrier_wait(&worker.scheduler.barrier)

  while not worker.scheduler.is_done:
    coroutine = worker_coro_get(worker)
    if coroutine:
      if has_sem:
        sem_post(&worker.scheduler.spin_loop_workers)
        has_sem = False
      was_waiting = False
      if coro_resume(coroutine) == coro_yield_value.MAY_RESUME:
        pthread_mutex_lock(&worker.mutex)
        worker.deque.push_back(coroutine)
        pthread_mutex_unlock(&worker.mutex)
      else:
        coro_free(coroutine)
    else:
      sem_post(&worker.scheduler.num_sleeping_workers)
      if was_waiting and not has_sem:
        sem_wait(&worker.scheduler.spin_loop_workers)
        # We initialized this semaphore to 0, so only
        # one waiting thread is not blocking on it.
        # ==> only one spin loop thread, others are parked.
        has_sem = True
      else:
        usleep(1000000) # 1 ms (arbitrary)
      was_waiting = True
      sem_wait(&worker.scheduler.num_sleeping_workers)

  if has_sem:
    sem_post(&worker.scheduler.spin_loop_workers)
    has_sem = False

  pthread_exit(NULL)

cdef coro_t *scheduler_coro_steal(worker_t *worker) nogil:
  cdef coro_t *coroutine = NULL
  cdef scheduler_t *scheduler = worker.scheduler

  pthread_mutex_lock(&scheduler.mutex)
  if not scheduler.deque.empty():
    coroutine = scheduler.deque.front()
    scheduler.deque.pop_front()
    coro_update(coroutine, &worker.switcher) # Update coroutine's data about its parent worker/thread.
  pthread_mutex_unlock(&scheduler.mutex)

  return coroutine

cdef coro_t *worker_coro_get(worker_t *worker) nogil:
  cdef coro_t *coroutine = NULL

  if not worker.bypass_global_queue:
    coroutine = scheduler_coro_steal(worker)

  worker.bypass_global_queue = worker.bypass_global_queue + 1 % 61

  if coroutine != NULL:
    return coroutine

  pthread_mutex_lock(&worker.mutex)
  if worker.deque.empty():
    pthread_mutex_unlock(&worker.mutex)
    coroutine = scheduler_coro_steal(worker)
    if coroutine == NULL:
      coroutine = worker_coro_steal(worker)
  else:
    coroutine = worker.deque.front()
    worker.deque.pop_front()
    pthread_mutex_unlock(&worker.mutex)

  return coroutine

cdef coro_t *worker_coro_steal(worker_t *thief) nogil:
  cdef:
    scheduler_t *scheduler = thief.scheduler
    worker_t *victim = NULL
    coro_t *stolen_coroutine = NULL
    unsigned short i, current_index
    unsigned short random_offset = rand() % scheduler.num_workers

  for i in range(scheduler.num_workers):
    current_index = (i + random_offset) % scheduler.num_workers
    victim = scheduler.workers + current_index

    if victim.deque == thief.deque:
      continue

    pthread_mutex_lock(&victim.mutex)
    if not victim.deque.empty():
      stolen_coroutine = victim.deque.back()
      victim.deque.pop_back()
      coro_update(stolen_coroutine, &thief.switcher) # Update coroutine's data about its parent worker/thread.
      pthread_mutex_unlock(&victim.mutex)
      return stolen_coroutine
    pthread_mutex_unlock(&victim.mutex)
  
  return NULL

cdef void worker_destroy(worker_t *worker) nogil:
  pthread_mutex_destroy(&worker.mutex)
  del worker.deque
